import os
import sys
from typing import Any, AsyncGenerator, cast

from streaq import Worker
from streaq.ui.deps import get_worker
from streaq.ui.tasks import router
from streaq.utils import import_string

__all__ = ["get_worker", "router"]


def run_web(host: str, port: int, worker_path: str) -> None:  # pragma: no cover
    import uvicorn
    from fastapi import FastAPI

    async def _get_worker() -> AsyncGenerator[Worker[Any], None]:
        sys.path.append(os.getcwd())
        worker = cast(Worker[Any], import_string(worker_path))
        yield worker

    app = FastAPI()
    app.dependency_overrides[get_worker] = _get_worker
    app.include_router(router)
    uvicorn.run(app, host=host, port=port)



from __future__ import annotations

from dataclasses import dataclass
from datetime import timedelta
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Generic,
    Optional,
    ParamSpec,
    Protocol,
    TypeAlias,
    TypeVar,
    overload,
)

if TYPE_CHECKING:  # pragma: no cover
    from streaq.task import RegisteredCron, RegisteredTask

P = ParamSpec("P")
POther = ParamSpec("POther")
R = TypeVar("R", bound=Optional[object])
ROther = TypeVar("ROther", bound=Optional[object])
WD = TypeVar("WD", bound=Optional[object])


@dataclass
class StreamMessage:
    """
    Dataclass wrapping data stored in the Redis stream.
    """

    message_id: str
    task_id: str
    priority: str


@dataclass
class TaskContext:
    """
    Dataclass containing task-specific information like the try count.
    """

    fn_name: str
    task_id: str
    timeout: timedelta | int | None
    tries: int
    ttl: timedelta | int | None


AnyCoroutine: TypeAlias = Coroutine[Any, Any, Any]
ReturnCoroutine: TypeAlias = Callable[..., AnyCoroutine]
TypedCoroutine: TypeAlias = Coroutine[Any, Any, R]

Middleware: TypeAlias = Callable[[ReturnCoroutine], ReturnCoroutine]

AsyncCron: TypeAlias = Callable[[], TypedCoroutine[R]]
SyncCron: TypeAlias = Callable[[], R]
AsyncTask: TypeAlias = Callable[P, TypedCoroutine[R]]
SyncTask: TypeAlias = Callable[P, R]


class CronDefinition(Protocol, Generic[WD]):
    @overload
    def __call__(self, fn: AsyncCron[R]) -> RegisteredCron[WD, R]: ...

    @overload
    def __call__(self, fn: SyncCron[R]) -> RegisteredCron[WD, R]: ...  # type: ignore


class TaskDefinition(Protocol, Generic[WD]):
    @overload
    def __call__(self, fn: AsyncTask[P, R]) -> RegisteredTask[WD, P, R]: ...

    @overload
    def __call__(self, fn: SyncTask[P, R]) -> RegisteredTask[WD, P, R]: ...  # type: ignore



from typing import Any, Generator
from uuid import uuid4

from pytest import fixture
from testcontainers.redis import RedisContainer

from streaq import Worker


@fixture(scope="module")
def anyio_backend() -> str:
    return "asyncio"


@fixture(scope="session")
def redis_container() -> Generator[RedisContainer, Any, None]:
    with RedisContainer() as container:
        yield container
        container.get_client().flushdb()


@fixture(scope="session")
def redis_url(redis_container: RedisContainer) -> Generator[str, None, None]:
    host = redis_container.get_container_host_ip()
    port = redis_container.get_exposed_port(redis_container.port)
    yield f"redis://{host}:{port}"


@fixture(scope="function")
def worker(redis_url: str) -> Worker:
    return Worker(redis_url=redis_url, queue_name=uuid4().hex)



import asyncio

import pytest
from anyio import create_task_group
from fastapi import FastAPI, HTTPException
from httpx import ASGITransport, AsyncClient

from streaq.ui import router
from streaq.ui.deps import get_worker
from streaq.worker import Worker

pytestmark = pytest.mark.anyio


async def test_no_override():
    with pytest.raises(HTTPException):
        _ = await get_worker()


async def test_get_pages(worker: Worker):
    app = FastAPI()
    prefix = "/streaq"
    app.include_router(router, prefix=prefix)
    worker.concurrency = 1
    worker.prefetch = 0

    @worker.task()
    async def sleeper(time: int) -> None:
        await asyncio.sleep(time)

    async def _get_worker():
        yield worker

    # queue up some tasks
    async with worker:
        scheduled = await sleeper.enqueue(10).start(delay=5)
        done = await sleeper.enqueue(0)
        running = await sleeper.enqueue(10)
        queued = await sleeper.enqueue(10)

    app.dependency_overrides[get_worker] = _get_worker
    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        await asyncio.sleep(2)
        async with AsyncClient(
            transport=ASGITransport(app=app), base_url="http://test"
        ) as client:
            # endpoints
            res = await client.get(f"{prefix}/")
            assert res.status_code == 303
            res = await client.get(f"{prefix}/queue")
            assert res.status_code == 200
            res = await client.patch(
                f"{prefix}/queue",
                data={
                    "functions": ["redis_health_check"],
                    "statuses": ["queued", "running", "done"],
                    "sort": "desc",
                },
            )
            assert res.status_code == 200

            res = await client.get(f"{prefix}/task/{done.id}")
            assert res.status_code == 200
            res = await client.get(f"{prefix}/task/{running.id}")
            assert res.status_code == 200
            res = await client.get(f"{prefix}/task/{queued.id}")
            assert res.status_code == 200
            res = await client.delete(f"{prefix}/task/{scheduled.id}")
            assert res.status_code == 200
            assert res.headers["HX-Redirect"] == f"{prefix}/queue"
        # cleanup worker
        tg.cancel_scope.cancel()



import pytest

from streaq.utils import import_string


def test_bad_path():
    with pytest.raises(ImportError):
        _ = import_string("asdf")


def test_bad_worker_name():
    with pytest.raises(ImportError):
        _ = import_string("example.asdf")



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from hashlib import sha256
from time import time
from typing import TYPE_CHECKING, Any, Generator, Generic
from uuid import UUID, uuid4

from anyio import fail_after
from crontab import CronTab

from streaq import logger
from streaq.constants import REDIS_TASK
from streaq.types import WD, AsyncCron, AsyncTask, P, POther, R, ROther
from streaq.utils import StreaqError, datetime_ms, now_ms, to_ms, to_seconds

if TYPE_CHECKING:  # pragma: no cover
    from streaq.worker import Worker


class StreaqRetry(StreaqError):
    """
    An exception you can manually raise in your tasks to make sure the task
    is retried.

    :param msg: error message to show
    :param delay:
        amount of time to wait before retrying the task; if none, will be the
        square of the number of attempts in seconds
    """

    def __init__(self, msg: str, delay: timedelta | int | None = None):
        super().__init__(msg)
        self.delay = delay


class TaskStatus(str, Enum):
    """
    Enum of possible task statuses:
    """

    NOT_FOUND = "missing"
    QUEUED = "queued"
    RUNNING = "running"
    SCHEDULED = "scheduled"
    DONE = "done"


@dataclass
class TaskInfo:
    """
    Dataclass containing information about a running or enqueued task.
    """

    fn_name: str
    enqueue_time: int
    tries: int
    scheduled: datetime | None
    dependencies: set[str]
    dependents: set[str]


@dataclass
class TaskResult(Generic[R]):
    """
    Dataclass wrapping the result of a task with additional information
    like run time and whether execution terminated successfully.
    """

    fn_name: str
    enqueue_time: int
    success: bool
    result: R | Exception
    start_time: int
    finish_time: int
    tries: int
    worker_id: str


@dataclass
class Task(Generic[R]):
    """
    Represents a task that has been enqueued or scheduled.

    Awaiting the object directly will enqueue it.
    """

    args: tuple[Any, ...]
    kwargs: dict[str, Any]
    parent: RegisteredCron[Any, R] | RegisteredTask[Any, Any, R]
    id: str = field(default_factory=lambda: uuid4().hex)
    _after: Task[Any] | None = None
    after: list[str] = field(default_factory=lambda: [])
    delay: timedelta | int | None = None
    schedule: datetime | None = None
    priority: str | None = None
    _triggers: Task[Any] | None = None

    def start(
        self,
        after: str | list[str] | None = None,
        delay: timedelta | int | None = None,
        schedule: datetime | None = None,
        priority: str | None = None,
    ) -> Task[R]:
        """
        Configure the task to modify schedule, queue, or dependencies.

        :param after: task ID(s) to wait for before running this task
        :param delay: duration to wait before running the task
        :param schedule: datetime at which to run the task
        :param priority: priority queue to insert the task

        :return: self
        """
        # merge _after and after into a single list[str]
        if isinstance(after, str):
            self.after.append(after)
        elif after:
            self.after.extend(after)
        self.delay = delay
        self.schedule = schedule
        self.priority = priority
        if (delay and schedule) or (delay and after) or (schedule and after):
            raise StreaqError(
                "Use one of 'delay', 'schedule', or 'after' when enqueuing tasks, not "
                "multiple!"
            )
        return self

    async def _enqueue(self) -> Task[R]:
        """
        This is called when the task is awaited.
        """
        if not self.parent.worker._stack:  # type: ignore
            raise StreaqError(
                "Worker did not initialize correctly, are you using the async context "
                "manager?"
            )
        if self._after:
            self.after.append(self._after.id)
        enqueue_time = now_ms()
        if self.schedule:
            score = datetime_ms(self.schedule)
        elif self.delay is not None:
            score = enqueue_time + to_ms(self.delay)
        else:
            score = 0
        data = self.serialize(enqueue_time)
        _priority = self.priority or self.parent.worker.priorities[-1]
        expire = to_ms(self.parent.expire or 0)
        if not await self.parent.worker.publish_task(
            keys=[
                self.parent.worker.stream_key,
                self.parent.worker.queue_key,
                self.task_key(REDIS_TASK),
                self.parent.worker.dependents_key,
                self.parent.worker.dependencies_key,
                self.parent.worker.results_key,
            ],
            args=[self.id, data, _priority, score, expire] + self.after,
        ):
            logger.debug("Task is unique and already exists, not enqueuing!")
        return self

    def then(
        self, task: RegisteredTask[WD, POther, ROther], **kwargs: Any
    ) -> Task[ROther]:
        """
        Enqueues the given task as a dependent of this one. Positional arguments will
        come from the previous task's output (tuple outputs will be unpacked), and any
        additional arguments can be passed as kwargs.

        :param task: task to feed output to

        :return: task object for newly created, dependent task
        """
        self._triggers = Task((), kwargs, task)
        self._triggers._after = self
        return self._triggers

    async def _chain(self) -> Task[R]:
        # traverse backwards
        if self._after:
            await self._after
        return await self._enqueue()

    def __await__(self) -> Generator[Any, None, Task[R]]:
        return self._chain().__await__()

    def task_key(self, mid: str) -> str:
        return self.parent.worker.prefix + mid + self.id

    def serialize(self, enqueue_time: int) -> Any:
        """
        Serializes the task data for sending to the queue.

        :param enqueue_time: the time at which the task was enqueued

        :return: serialized task data
        """
        try:
            data = {
                "f": self.parent.fn_name,
                "a": self.args,
                "k": self.kwargs,
                "t": enqueue_time,
            }
            if self._after:
                data["A"] = self._after.id
            if self._triggers:
                data["T"] = self._triggers.id
            return self.parent.worker.serialize(data)
        except Exception as e:
            raise StreaqError(f"Unable to serialize task {self.parent.fn_name}!") from e

    async def status(self) -> TaskStatus:
        """
        Fetch the current status of the task.

        :return: current task status
        """
        return await self.parent.worker.status_by_id(self.id)

    async def result(self, timeout: timedelta | int | None = None) -> TaskResult[R]:
        """
        Wait for and return the task's result, optionally with a timeout.

        :param timeout: amount of time to wait before raising a `TimeoutError`

        :return: wrapped result object
        """
        return await self.parent.worker.result_by_id(self.id, timeout=timeout)

    async def abort(self, timeout: timedelta | int = 5) -> bool:
        """
        Notify workers that the task should be aborted.

        :param timeout: how long to wait to confirm abortion was successful

        :return: whether the task was aborted successfully
        """
        return await self.parent.worker.abort_by_id(self.id, timeout=timeout)

    async def info(self) -> TaskInfo | None:
        """
        Fetch info about a previously enqueued task.

        :return: task info, unless task has finished or doesn't exist
        """
        return await self.parent.worker.info_by_id(self.id)


@dataclass
class RegisteredTask(Generic[WD, P, R]):
    fn: AsyncTask[P, R]
    expire: timedelta | int | None
    max_tries: int | None
    silent: bool
    timeout: timedelta | int | None
    ttl: timedelta | int | None
    unique: bool
    worker: Worker[WD]
    _fn_name: str | None = None

    def enqueue(
        self,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Task[R]:
        """
        Serialize the task and send it to the queue for later execution by an
        active worker. Though this isn't async, it should be awaited as it
        returns an object that should be.
        """
        return Task(args, kwargs, self)

    async def run(self, *args: P.args, **kwargs: P.kwargs) -> R:
        """
        Run the task in the local event loop with the given params and return the
        result. This skips enqueuing and result storing in Redis.
        """
        with fail_after(to_seconds(self.timeout)):
            return await self.fn(*args, **kwargs)

    @property
    def fn_name(self) -> str:
        return self._fn_name or self.fn.__qualname__


@dataclass
class RegisteredCron(Generic[WD, R]):
    fn: AsyncCron[R]
    crontab: CronTab
    max_tries: int | None
    silent: bool
    timeout: timedelta | int | None
    ttl: timedelta | int | None
    unique: bool
    worker: Worker[WD]
    expire: timedelta | int | None = None
    _fn_name: str | None = None

    def enqueue(self) -> Task[R]:
        """
        Serialize the task and send it to the queue for later execution by an
        active worker. Though this isn't async, it should be awaited as it
        returns an object that should be.
        """
        task = Task((), {}, self)
        uid_bytes = f"{self.fn_name}@{datetime_ms(self.schedule())}".encode()
        deterministic_hash = sha256(uid_bytes).hexdigest()
        task.id = UUID(bytes=bytes.fromhex(deterministic_hash[:32]), version=4).hex
        return task

    async def run(self) -> R:
        """
        Run the task in the local event loop and return the result.
        This skips enqueuing and result storing in Redis.
        """
        with fail_after(to_seconds(self.timeout)):
            return await self.fn()

    @property
    def fn_name(self) -> str:
        return self._fn_name or self.fn.__qualname__

    def schedule(self) -> datetime:
        """
        Datetime of next run.
        """
        return datetime.fromtimestamp(self.next() / 1000, tz=self.worker.tz)

    def next(self) -> int:
        """
        Timestamp in milliseconds of next run.
        """
        return round((time() + self.delay) * 1000)

    @property
    def delay(self) -> float:
        return self.crontab.next(now=datetime.now(self.worker.tz))  # type: ignore



import asyncio
import time
from datetime import datetime, timedelta
from typing import Any
from uuid import uuid4

import pytest
from anyio import create_task_group

from streaq import StreaqError, Worker
from streaq.constants import REDIS_UNIQUE
from streaq.task import StreaqRetry, TaskStatus
from streaq.types import ReturnCoroutine

pytestmark = pytest.mark.anyio


async def test_result_timeout(worker: Worker):
    @worker.task()
    async def foobar() -> None:
        await asyncio.sleep(5)

    async with worker:
        task = await foobar.enqueue()

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        with pytest.raises(TimeoutError):
            await task.result(3)
        tg.cancel_scope.cancel()


async def test_task_timeout(worker: Worker):
    @worker.task(timeout=1)
    async def foobar() -> None:
        await asyncio.sleep(5)

    async with worker:
        task = await foobar.enqueue()

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        res = await task.result(3)
        assert not res.success
        assert isinstance(res.result, TimeoutError)
        tg.cancel_scope.cancel()


async def test_task_status(worker: Worker):
    @worker.task()
    async def foobar() -> None:
        await asyncio.sleep(1)

    async with worker:
        task = foobar.enqueue()
        assert await task.status() == TaskStatus.NOT_FOUND
        await task.start()
        assert await task.status() == TaskStatus.QUEUED
        task2 = await foobar.enqueue().start(delay=5)

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        await task.result(3)
        assert await task.status() == TaskStatus.DONE
        assert await task2.status() == TaskStatus.SCHEDULED
        tg.cancel_scope.cancel()


async def test_task_status_running(worker: Worker):
    @worker.task()
    async def foobar() -> None:
        await asyncio.sleep(3)

    async with worker:
        task = await foobar.enqueue().start()

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        await asyncio.sleep(1)
        assert await task.status() == TaskStatus.RUNNING
        tg.cancel_scope.cancel()


async def test_task_cron(worker: Worker):
    @worker.cron("30 9 1 1 *")
    async def cron1() -> bool:
        return True

    @worker.cron("* * * * * * *")  # once/second
    async def cron2() -> None:
        await asyncio.sleep(3)

    schedule = cron1.schedule()
    assert schedule.day == 1 and schedule.month == 1
    assert await cron1.run()
    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        await asyncio.sleep(2)
        # this will be set if task is running
        assert await worker.redis.get(worker.prefix + REDIS_UNIQUE + cron2.fn_name)
        tg.cancel_scope.cancel()


async def test_task_info(worker: Worker):
    @worker.task()
    async def foobar() -> None:
        pass

    async with worker:
        task = await foobar.enqueue().start(delay=5)
        task2 = await foobar.enqueue()
        info = await task.info()
        info2 = await task2.info()
        assert info and info.scheduled is not None
        assert info2 and info2.scheduled is None


async def test_task_retry(worker: Worker):
    @worker.task()
    async def foobar() -> int:
        ctx = worker.task_context()
        if ctx.tries < 3:
            raise StreaqRetry("Retrying!")
        return ctx.tries

    async with worker:
        task = await foobar.enqueue()

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        res = await task.result(7)
        assert res.success
        assert res.result == 3
        tg.cancel_scope.cancel()


async def test_task_retry_with_delay(worker: Worker):
    @worker.task()
    async def foobar() -> int:
        ctx = worker.task_context()
        if ctx.tries == 1:
            raise StreaqRetry("Retrying!", delay=timedelta(seconds=3))
        return ctx.tries

    async with worker:
        task = await foobar.enqueue()

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        with pytest.raises(TimeoutError):
            await task.result(3)
        res = await task.result(1)
        assert res is not None
        assert res.success
        assert res.result == 2
        tg.cancel_scope.cancel()


async def test_task_failure(worker: Worker):
    @worker.task()
    async def foobar() -> None:
        raise Exception("That wasn't supposed to happen!")

    async with worker:
        task = await foobar.enqueue()

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        res = await task.result(3)
        assert not res.success
        assert isinstance(res.result, Exception)
        tg.cancel_scope.cancel()


async def test_task_retry_no_delay(worker: Worker):
    @worker.task()
    async def foobar() -> bool:
        if worker.task_context().tries == 1:
            raise StreaqRetry("Retrying!", delay=0)
        return True

    async with worker:
        task = await foobar.enqueue()

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        res = await task.result(3)
        assert res is not None
        assert res.success
        assert res.result
        tg.cancel_scope.cancel()


async def test_task_max_retries(worker: Worker):
    @worker.task()
    async def foobar() -> None:
        raise StreaqRetry("Retrying!", delay=0)

    async with worker:
        task = await foobar.enqueue()

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        res = await task.result(3)
        assert res is not None
        assert not res.success
        assert isinstance(res.result, StreaqError)
        tg.cancel_scope.cancel()


async def test_task_failed_abort(worker: Worker):
    @worker.task()
    async def foobar() -> bool:
        return True

    worker.burst = True
    async with worker:
        task = await foobar.enqueue()
    await worker.run_async()
    result = await task.result(3)
    assert result.success
    assert result.result
    assert not await task.abort()


async def test_task_nonexistent_or_finished_dependency(worker: Worker):
    @worker.task()
    async def foobar() -> None:
        pass

    async with worker:
        task = await foobar.enqueue().start(after="nonexistent")
    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        with pytest.raises(TimeoutError):
            await task.result(3)
        tg.cancel_scope.cancel()


async def test_task_dependency(worker: Worker):
    @worker.task()
    async def foobar() -> None:
        await asyncio.sleep(1)

    async with worker:
        task = await foobar.enqueue().start(delay=1)
        task2 = await foobar.enqueue().start(after=task.id)

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        assert await task2.status() == TaskStatus.SCHEDULED
        await task.result(3)
        result = await task2.result(3)
        assert result.success
        tg.cancel_scope.cancel()


async def test_task_dependency_multiple(worker: Worker):
    @worker.task()
    async def foobar() -> None:
        await asyncio.sleep(1)

    async with worker:
        task = await foobar.enqueue().start()
        task2 = await foobar.enqueue().start(after=task.id)
        task3 = await foobar.enqueue().start(after=[task.id, task2.id])

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        assert await task2.status() == TaskStatus.SCHEDULED
        assert await task3.status() == TaskStatus.SCHEDULED
        res1 = await task.result(3)
        assert res1.success
        assert await task3.status() == TaskStatus.SCHEDULED
        res2 = await task2.result(3)
        assert res2.success
        res3 = await task3.result(3)
        assert res3.success
        tg.cancel_scope.cancel()


async def test_task_dependency_failed(worker: Worker):
    @worker.task()
    async def foobar() -> None:
        raise Exception("Oh no!")

    @worker.task()
    async def do_nothing() -> None:
        pass

    async with worker:
        task = await foobar.enqueue().start()
        dep = await do_nothing.enqueue().start(after=task.id)

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        res = await dep.result(3)
        assert not res.success
        assert isinstance(res.result, StreaqError)
        tg.cancel_scope.cancel()


async def test_sync_task(worker: Worker):
    @worker.task()
    def foobar() -> None:
        time.sleep(2)

    async with worker:
        task = await foobar.enqueue()
        task2 = await foobar.enqueue()

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        # this would time out if these were running sequentially
        results = await asyncio.gather(task.result(3), task2.result(3))
        assert all(res.success for res in results)
        tg.cancel_scope.cancel()


async def test_unsafe_enqueue(worker: Worker):
    @worker.task()
    async def foobar(ret: int) -> int:
        return ret

    async with worker:
        task = await worker.enqueue_unsafe(foobar.fn_name, 42)

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        res = await task.result(3)
        assert res.success
        assert res.result == 42
        tg.cancel_scope.cancel()


async def test_chained_failed_dependencies(worker: Worker):
    @worker.task()
    async def foobar() -> None:
        raise Exception("Oh no!")

    @worker.task()
    async def child() -> None:
        pass

    async with worker:
        task = await foobar.enqueue().start(delay=timedelta(seconds=3))
        dep1 = await child.enqueue().start(after=task.id)
        dep2 = await child.enqueue().start(after=[task.id, dep1.id])

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        await asyncio.sleep(1)
        assert await task.abort(3)
        res1 = await dep1.result(3)
        res2 = await dep2.result(3)
        assert not res1.success and isinstance(res1.result, StreaqError)
        assert not res2.success and isinstance(res2.result, StreaqError)
        tg.cancel_scope.cancel()


async def test_task_priorities(redis_url: str):
    worker = Worker(
        redis_url=redis_url,
        queue_name=uuid4().hex,
        concurrency=4,
        priorities=["low", "high"],
    )

    @worker.task()
    async def foobar() -> None:
        await asyncio.sleep(1)

    async with worker:
        low = [foobar.enqueue().start(priority="low") for _ in range(4)]
        high = [foobar.enqueue().start(priority="high") for _ in range(4)]
        await worker.enqueue_many(low + high)

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        results = await asyncio.gather(*[t.result(3) for t in high])
        statuses = await asyncio.gather(*[t.status() for t in low])
        assert all(res.success for res in results)
        assert all(status != TaskStatus.DONE for status in statuses)
        tg.cancel_scope.cancel()


async def test_scheduled_task(worker: Worker):
    @worker.task()
    async def foobar() -> None:
        pass

    dt = datetime.now() + timedelta(seconds=1)
    async with worker:
        task = await foobar.enqueue().start(schedule=dt)

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        assert await task.status() == TaskStatus.SCHEDULED
        res = await task.result(3)
        assert res.success
        tg.cancel_scope.cancel()


async def test_bad_start_params(worker: Worker):
    @worker.task()
    async def foobar() -> None:
        pass

    with pytest.raises(StreaqError):
        await foobar.enqueue().start(delay=1, schedule=datetime.now())
    with pytest.raises(StreaqError):
        await foobar.enqueue().start(delay=1, after="foobar")
    with pytest.raises(StreaqError):
        await foobar.enqueue().start(schedule=datetime.now(), after="foobar")


async def test_enqueue_unique_task(worker: Worker):
    @worker.task(unique=True)
    async def foobar() -> None:
        await asyncio.sleep(1)

    async with worker:
        task = await foobar.enqueue()
        task2 = await foobar.enqueue()

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        results = await asyncio.gather(task.result(), task2.result())
        assert any(isinstance(r.result, StreaqError) for r in results)
        assert any(r.result is None for r in results)
        tg.cancel_scope.cancel()


async def test_failed_abort(worker: Worker):
    @worker.task(ttl=0)
    async def foobar() -> None:
        pass

    async with worker:
        task = await foobar.enqueue().start()

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        await asyncio.sleep(1)
        assert not await task.abort(1)
        tg.cancel_scope.cancel()


async def test_cron_run(worker: Worker):
    @worker.cron("* * * * * * *")
    async def cron1() -> bool:
        return True

    @worker.cron("* * * * * * *", timeout=1)
    async def cron2() -> None:
        await asyncio.sleep(3)

    async with worker:
        assert await cron1.run()
        with pytest.raises(TimeoutError):
            await cron2.run()


async def test_sync_cron(worker: Worker):
    @worker.cron("* * * * * * *")
    def cronjob() -> None:
        time.sleep(3)

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        await asyncio.sleep(2)
        assert await worker.redis.get(worker.prefix + REDIS_UNIQUE + cronjob.fn_name)
        tg.cancel_scope.cancel()


async def test_cron_multiple_runs(worker: Worker):
    val = 0

    @worker.cron("* * * * * * *")
    async def cronjob() -> None:
        nonlocal val
        val += 1

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        await asyncio.sleep(5)
        assert val > 1
        tg.cancel_scope.cancel()


async def test_middleware(worker: Worker):
    @worker.task()
    async def foobar() -> int:
        return 2

    @worker.middleware
    def double(task: ReturnCoroutine) -> ReturnCoroutine:
        async def wrapper(*args, **kwargs) -> Any:
            result = await task(*args, **kwargs)
            return result * 2

        return wrapper

    async with worker:
        task = await foobar.enqueue()

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        res = await task.result(3)
        assert res.success
        assert res.result == 4
        tg.cancel_scope.cancel()


async def test_task_pipeline(worker: Worker):
    @worker.task()
    async def double(val: int) -> int:
        return val * 2

    @worker.task()
    async def is_even(val: int) -> bool:
        return val % 2 == 0

    async with worker:
        task = await double.enqueue(1).then(double).then(is_even)

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        res = await task.result(3)
        assert res.result and res.success
        tg.cancel_scope.cancel()


async def test_task_with_custom_name(worker: Worker):
    @worker.task(name="bar")
    async def foo() -> int:
        return 42

    async def foobar():
        return

    assert foo.fn_name == "bar"
    with pytest.raises(StreaqError):
        worker.task(name="bar")(foobar)

    @worker.task()
    async def bar():
        return 10

    async with worker:
        task1 = await worker.enqueue_unsafe("bar")
        task2 = await worker.enqueue_unsafe(bar.fn_name)
        task3 = await worker.enqueue_unsafe(foo.fn.__qualname__)

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        res = await task1.result(3)
        assert res.result == 42
        res = await task2.result(3)
        assert res.result == 10
        res = await task3.result(3)
        assert not res.success
        tg.cancel_scope.cancel()


async def test_cron_with_custom_name(worker: Worker):
    @worker.cron("* * * * * * *", name="foo")
    async def cronjob() -> None:
        await asyncio.sleep(3)

    async def cronjob1() -> None:
        pass

    assert cronjob.fn_name == "foo"
    with pytest.raises(StreaqError):
        worker.cron("* * * * * * *", name="foo")(cronjob1)

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        await asyncio.sleep(2)
        assert await worker.redis.get(worker.prefix + REDIS_UNIQUE + cronjob.fn_name)
        tg.cancel_scope.cancel()


@pytest.mark.parametrize("ttl", [60, 0])
@pytest.mark.parametrize("wait", [1, 0])
async def test_abort(worker: Worker, ttl: int, wait: int):
    @worker.task(ttl=ttl)
    async def foobar() -> None:
        await asyncio.sleep(5)

    async with worker:
        task = await foobar.enqueue()

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        await asyncio.sleep(wait)
        assert await task.abort(3)
        tg.cancel_scope.cancel()


async def test_task_expired(worker: Worker):
    @worker.task(expire=1)
    async def foobar() -> None:
        pass

    async with worker:
        task = await foobar.enqueue()
        await asyncio.sleep(1)

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        res = await task.result(3)
        assert not res.success and isinstance(res.result, StreaqError)
        tg.cancel_scope.cancel()



import subprocess
import sys

import pytest
from typer.testing import CliRunner

from streaq import VERSION, Worker
from streaq.cli import cli

pytestmark = pytest.mark.anyio
runner = CliRunner()
test_module = sys.modules["tests.test_cli"]


def test_burst(worker: Worker):
    setattr(test_module, "test_worker", worker)
    result = runner.invoke(cli, ["tests.test_cli.test_worker", "--burst"])
    assert result.exit_code == 0


def test_multiple_workers(worker: Worker):
    setattr(test_module, "test_worker", worker)
    result = runner.invoke(
        cli, ["--burst", "--workers", "2", "tests.test_cli.test_worker"]
    )
    assert result.exit_code == 0


def test_verbose(worker: Worker):
    setattr(test_module, "test_worker", worker)
    result = runner.invoke(cli, ["tests.test_cli.test_worker", "--burst", "--verbose"])
    assert result.exit_code == 0
    assert "enqueuing" in result.stderr


def test_version(worker: Worker):
    setattr(test_module, "test_worker", worker)
    result = runner.invoke(cli, ["--version"])
    assert result.exit_code == 0
    assert VERSION in result.stdout


def test_help(worker: Worker):
    setattr(test_module, "test_worker", worker)
    result = runner.invoke(cli, ["--help"])
    assert result.exit_code == 0
    assert "--help" in result.stdout


def test_main_entry_point():
    result = subprocess.run(
        [sys.executable, "-m", "streaq", "--help"],
        capture_output=True,
        text=True,
        check=False,
    )
    assert result.returncode == 0
    assert "--help" in result.stdout


"""
@pytest.fixture(scope="function")
def worker_file(redis_url: str, tmp_path: Path):
    tmp = NamedTemporaryFile(dir=tmp_path, suffix=".py", mode="w+", delete=False)
    tmp.write(f""from uuid import uuid4
from streaq import Worker
worker = Worker(redis_url="{redis_url}", queue_name=uuid4().hex)"")
    tmp.close()
    yield tmp.name
    os.remove(tmp.name)


async def test_watch(worker_file: str, tmp_path: Path):
    file_name = worker_file.split("/")[-1][:-3]

    def run_subprocess():
        with pytest.raises(subprocess.TimeoutExpired) as e:
            subprocess.run(
                [sys.executable, "-m", "streaq", f"{file_name}.worker", "--reload"],
                capture_output=True,
                text=True,
                check=False,
                timeout=3,
                cwd=tmp_path,
            )
        return e

    async def modify_file():
        await asyncio.sleep(1)  # wait for startup
        with open(worker_file, "a") as f:
            f.write("  # change from test")

    res, _ = await asyncio.gather(asyncio.to_thread(run_subprocess), modify_file())
    assert str(res.value.stderr).count("starting") > 1


def find_free_port() -> int:  # Finds and returns an available TCP port.
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("", 0))
        return s.getsockname()[1]


async def test_web_cli(worker_file: str):
    file_name = worker_file.split("/")[-1][:-3]

    def run_subprocess():
        with pytest.raises(subprocess.TimeoutExpired) as e:
            subprocess.run(
                [
                    sys.executable,
                    "-m",
                    "streaq",
                    f"{file_name}.worker",
                    "--web",
                    "--port",
                    str(find_free_port()),
                ],
                capture_output=True,
                text=True,
                check=False,
                timeout=3,
            )
        return e

    async def modify_file():
        await asyncio.sleep(1)  # wait for startup
        return httpx.get("http://localhost:8000/")

    web, res = await asyncio.gather(asyncio.to_thread(run_subprocess), modify_file())
    assert "Uvicorn" in str(web.value.stderr)
    assert res.status_code == 303
"""



import sys

import pytest
from anyio import run, sleep

from streaq import Worker


@pytest.mark.anyio
async def test_reclaim_idle_task(redis_url: str):
    worker1 = Worker(redis_url=redis_url, queue_name="reclaim")

    @worker1.task(timeout=3)
    async def foo() -> None:
        await sleep(2)

    await worker1.run_async()


if __name__ == "__main__":
    redis_url = sys.argv[1]
    run(test_reclaim_idle_task, redis_url)



import asyncio
import json
import os
import pickle
import secrets
import signal
import subprocess
import sys
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import Any, AsyncIterator
from uuid import uuid4

import pytest
from anyio import create_task_group, move_on_after

from streaq.constants import REDIS_TASK
from streaq.task import TaskStatus
from streaq.utils import StreaqError
from streaq.worker import Worker

NAME_STR = "Freddy"
pytestmark = pytest.mark.anyio


async def test_worker_redis(redis_url: str):
    worker = Worker(redis_url=redis_url, queue_name=uuid4().hex)
    await worker.redis.ping()


@dataclass
class WorkerContext:
    name: str


@asynccontextmanager
async def deps() -> AsyncIterator[WorkerContext]:
    yield WorkerContext(NAME_STR)


async def test_lifespan(redis_url: str):
    worker = Worker(redis_url=redis_url, lifespan=deps, queue_name=uuid4().hex)

    @worker.task()
    async def foobar() -> str:
        return worker.context.name

    @worker.task(timeout=1)
    async def foobar2() -> None:
        await asyncio.sleep(3)

    async with worker:
        res = await foobar.run()
        assert res == NAME_STR
        with pytest.raises(TimeoutError):
            await foobar2.run()


async def test_health_check(redis_url: str):
    worker = Worker(
        redis_url=redis_url,
        health_crontab="* * * * * * *",
        queue_name=uuid4().hex,
    )
    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        await asyncio.sleep(2)
        worker_health = await worker.redis.get(f"{worker._health_key}:{worker.id}")
        redis_health = await worker.redis.get(worker._health_key + ":redis")
        assert worker_health is not None
        assert redis_health is not None
        tg.cancel_scope.cancel()


async def test_queue_size(worker: Worker):
    assert await worker.queue_size() == 0


def raise_error(*arg, **kwargs) -> Any:
    raise Exception("Couldn't serialize/deserialize!")


async def test_bad_serializer(redis_url: str):
    worker = Worker(redis_url=redis_url, serializer=raise_error, queue_name=uuid4().hex)

    @worker.task()
    async def foobar() -> None:
        print("This can't print!")

    async with worker:
        with pytest.raises(StreaqError):
            await foobar.enqueue()


async def test_bad_deserializer(redis_url: str):
    worker = Worker(
        redis_url=redis_url, deserializer=raise_error, queue_name=uuid4().hex
    )

    @worker.task()
    async def foobar() -> None:
        print("This can't print!")

    worker.burst = True
    async with worker:
        task = await foobar.enqueue()

    await worker.run_async()
    with pytest.raises(StreaqError):
        await task.result(3)


async def test_custom_serializer(worker: Worker):
    worker.serializer = json.dumps
    worker.deserializer = json.loads

    @worker.task()
    async def foobar() -> None:
        pass

    async with worker:
        task = await foobar.enqueue()

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        assert (await task.result(3)).success
        tg.cancel_scope.cancel()


async def test_uninitialized_worker(worker: Worker):
    @worker.task()
    async def foobar() -> None:
        print(worker.context)

    with pytest.raises(StreaqError):
        await foobar.run()
    with pytest.raises(StreaqError):
        await foobar.enqueue()


async def test_active_tasks(worker: Worker):
    @worker.task()
    async def foo() -> None:
        await asyncio.sleep(3)

    n_tasks = 5
    async with worker:
        tasks = [foo.enqueue() for _ in range(n_tasks)]
        await worker.enqueue_many(tasks)

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        await asyncio.sleep(1)
        assert worker.active == n_tasks
        tg.cancel_scope.cancel()


async def test_handle_signal(worker: Worker):
    @worker.task()
    async def foo() -> None:
        await asyncio.sleep(3)

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        await asyncio.sleep(1)
        os.kill(os.getpid(), signal.SIGINT)

    async with worker:
        task = await foo.enqueue()
        assert await task.status() == TaskStatus.QUEUED


async def test_reclaim_backed_up(redis_url: str):
    queue_name = uuid4().hex
    worker = Worker(
        concurrency=2, redis_url=redis_url, queue_name=queue_name, idle_timeout=2
    )
    worker2 = Worker(redis_url=redis_url, queue_name=queue_name)

    async def foo() -> None:
        await asyncio.sleep(3)

    registered = worker.task()(foo)
    worker2.task()(foo)

    # enqueue tasks
    async with worker:
        tasks = [registered.enqueue() for _ in range(4)]
        await worker.enqueue_many(tasks)
    async with create_task_group() as tg:
        # run first worker which will pick up all tasks
        tg.start_soon(worker.run_async)
        await asyncio.sleep(1)
        # run second worker which will pick up prefetched tasks
        tg.start_soon(worker2.run_async)

        results = await asyncio.gather(*[t.result(8) for t in tasks])
        assert any(r.worker_id == worker2.id for r in results)
        tg.cancel_scope.cancel()


async def test_reclaim_idle_task(redis_url: str):
    worker2 = Worker(redis_url=redis_url, queue_name="reclaim")

    @worker2.task(timeout=3)
    async def foo() -> None:
        await asyncio.sleep(2)

    # enqueue task
    async with worker2:
        task = await foo.enqueue()
    # run separate worker which will pick up task
    worker = subprocess.Popen([sys.executable, "tests/failure.py", redis_url])
    await asyncio.sleep(1)
    # kill worker abruptly to disallow cleanup
    os.kill(worker.pid, signal.SIGKILL)
    worker.wait()

    async with create_task_group() as tg:
        tg.start_soon(worker2.run_async)
        assert (await task.result(8)).success
        tg.cancel_scope.cancel()


async def test_change_cron_schedule(redis_url: str):
    async def foo() -> None:
        pass

    worker = Worker(redis_url=redis_url, queue_name=uuid4().hex)
    foo1 = worker.cron("0 0 1 1 *")(foo)
    with move_on_after(2):
        await worker.run_async()
    task1 = foo1.enqueue()
    info = await task1.info()
    assert info and foo1.schedule() == info.scheduled

    worker2 = Worker(redis_url=redis_url, queue_name=worker.queue_name)
    foo2 = worker2.cron("1 0 1 1 *")(foo)  # 1 minute later
    async with create_task_group() as tg:
        tg.start_soon(worker2.run_async)
        await asyncio.sleep(2)
        tg.cancel_scope.cancel()
    task2 = foo2.enqueue()
    info2 = await task2.info()
    assert info2 and foo2.schedule() == info2.scheduled
    assert foo1.schedule() != foo2.schedule()


async def test_signed_data(redis_url: str):
    worker = Worker(
        redis_url=redis_url,
        queue_name=uuid4().hex,
        signing_secret=secrets.token_urlsafe(32),
    )

    @worker.task()
    async def foo() -> str:
        return "bar"

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        async with worker:
            task = await foo.enqueue()
            res = await task.result(3)
            assert res.success and res.result == "bar"
        tg.cancel_scope.cancel()


async def test_sign_non_binary_data(redis_url: str):
    worker = Worker(
        redis_url=redis_url,
        queue_name=uuid4().hex,
        signing_secret=secrets.token_urlsafe(32),
        serializer=json.dumps,
    )

    @worker.task()
    async def foo() -> str:
        return "bar"

    async with worker:
        with pytest.raises(StreaqError):
            await foo.enqueue()


async def test_corrupt_signed_data(redis_url: str):
    worker = Worker(
        redis_url=redis_url,
        queue_name=uuid4().hex,
        handle_signals=False,
        signing_secret=secrets.token_urlsafe(32),
    )

    @worker.task()
    async def foo() -> str:
        return "bar"

    async with worker:
        task = await foo.enqueue()
        await worker.redis.set(
            task.task_key(REDIS_TASK), pickle.dumps({"f": "This is an attack!"})
        )

    async with create_task_group() as tg:
        tg.start_soon(worker.run_async)
        res = await task.result(5)
        assert not res.success and isinstance(res.result, StreaqError)
        tg.cancel_scope.cancel()


async def test_enqueue_many(worker: Worker):
    @worker.task()
    async def foobar(val: int) -> int:
        await asyncio.sleep(1)
        return val

    async with worker:
        tasks = [foobar.enqueue(i) for i in range(10)]
        await worker.enqueue_many(tasks)
    assert await worker.queue_size() >= len(tasks)


async def test_invalid_task_context(worker: Worker):
    with pytest.raises(StreaqError):
        worker.task_context()
